
/**
 * Define a classe SacoMoedas e cdigo de teste associado.
 * Cada objecto desta classe representa um saco de moedas (euros ou cntimos
 * de euro), ou, mais precisamente, os valores dessas moedas.
 * Todos os valores so em cntimos de euro.
 * 
 * @author (Joo Pascoal Faria) 
 * @version (24 de Fevereiro de 2003)
 * Nota: Inclui verso melhorada de "transferir".
 */
public class SacoMoedas
{
    /**
     * Array onde se definem os valores das moedas possveis, em cntimos 
     * de euro (e no em euros, para evitar trabalhar com nmeros em 
     * vrgula flutuante, com os consequentes erros de arredondamento).
     */
    public static final int moedas[] = {1, 2, 5, 10, 20, 50, 100, 200};
    
    /**
     * Array usado internamente para guardar a quantidade de moedas 
     * existente no saco para cada um dos valores possiveis. Mais 
     * precisamente, "quant[i]"  o nmero de moedas existentes no 
     * saco de valor "moedas[i]".
     */
    private int quant[];
    
    /**
     * Construtor. Inicializa as quantidades a 0.
     */
    public SacoMoedas()
    {
        quant = new int[moedas.length];
    }

    /**
     * Indica se o saco aceita moedas com o valor indicado no argumento.
     * Nesta classe retorna sempre "true", mas pode ser redefindo 
     * (overridden) pelas subclasses.
     */
    public boolean aceitaMoeda(int valor)
    {
        return true;
    }
             
    /**
     * Mtodo usado internamente.
     * Retorna o ndice no array "moedas" do valor passado como argumento.
     * No caso de no encontrar, retorna -1.
     */
    private static int getIndice(int valor)
    {
        for (int i = 0; i < moedas.length; i++)
            if (moedas[i] == valor)
                return i;
        return -1;
    }
        
    /**
     * Retorna a quantidade de moedas existentes no saco de valor igual 
     * ao indicado no argumento. 
     * Usa o mtodo "getIndice".
     */
    public int getQuant(int valor)
    {
        int i = getIndice(valor);
        if (i == -1)
            return 0;
        else
            return quant[i];
    }

    /**
     * Estabelece a quantidade de moedas existentes no saco de valor igual 
     * ao indicado no argumento.
     * Se o saco no aceitar moedas com o valor indicado (porque o valor no
     * est definido no array "moedas" ou o mtodo "aceitaMoeda" devolve 
     * falso), no faz nada.
     * Usa o mtodo "getIndice".
     */
    public void setQuant(int valor, int qt)
    {
        int i = getIndice(valor);
        if (i == -1)
            return;
        if ( ! aceitaMoeda(valor))
            return;
        quant[i] = qt;
    }
  
    /**
     * Retorna o montante (valor) total das moedas existentes no saco.
     */
    public int getMontante()
    {
        int sum = 0;
        for (int i = moedas.length-1; i >= 0; i--)
            sum += quant[i] * moedas[i];
        return sum;
    }
    
    /**
     * Retorna uma string representando o conteudo do saco.
     * Por exemplo, se o saco contiver duas moedas de valor 1 (cntimo) e
     * uma moeda de valor 20 (cntimos), retorna a string "{1, 1, 20}".
     */
    public String toString()
    {
        String s = "{";
        boolean first = true;
        
        for (int i = 0; i < moedas.length; i++)
            for (int j = 0; j < quant[i]; j++)
            {
                if (first) 
                {
                    s = s + moedas[i];
                    first = false;
                }
                else
                    s = s + ", " + moedas[i];
            }
        
        s = s + "}";
        return s;
    }
 
    /**
     * Tenta transferir um dado montante de moedas deste saco (origem) 
     * para um saco de destino. 
     * A transferncia no  possvel quando o montante existente no saco 
     * de origem  inferior ao pretendido ou quando o saco de origem no 
     * tem moedas que permitam perfazer exactamente o montante pretendido.
     * Por exemplo, no  possivel retirar 1 cntimo de um saco que s tem 
     * uma moeda de 20 cntimos.
     * No caso de consguir transferir, retorna "true". Seno, retorna 
     * "false".
     * 
     * Sugere-se o seguinte algoritmo de "fora bruta":
     *  1. Se o montante pretendido for 0, terminar com sucesso.
     *  2. Escolher a moeda de maior valor do saco de origem que no excede 
     *     o montante pedido e pode ser aceite pelo saco de destino (ao 
     *     escolher a moeda de maior valor tende-se a minimizar o nmero de
     *     moedas transferidas).
     *  3. Se no existir nenhuma moeda nessas condies, terminar com 
     *     indicao de insucesso.
     *  4. Existindo uma moeda nessas condies, 
     *     4.1. Transferir essa moeda para o saco de destino.
     *     4.2. Tentar transferir o montante restante invocando o mesmo
     *          mtodo recursivamente.
     *     4.3. Se foi possvel transferir o montante restante, terminar 
     *          com indicao de sucesso.
     *     4.4. Se no foi possvel transferir o montante restante,
     *          4.5. Repor a situao anterior, transferindo a moeda de 
     *               volta para o saco de origem
     *          4.6. Voltar ao passo 2, escolhendo agora uma moeda de 
     *               valor inferior
     * 
     * Numa segunda fase, sugerem-se os seguintes melhoramentos 
     * (valorizao para alm de 20 valores):
     * 1) Modificar o algoritmo para garantir que de facto  transferido um 
     *    nmero mnimo de moedas. Por exemplo, ao tentar tirar 60 cntimos
     *    de um saco com as moedas {2, 2, 2, 2, 2, 20, 20, 20, 50} deve-se
     *    tirar {20, 20, 20} e no {2, 2, 2, 2, 2, 50}.
     * 2) Modificar o algoritmo para, se necessrio, tranferir troco do 
     *    saco de destino para o saco de origem. Por exemplo, para 
     *    transferir 4 cntimos de um saco com uma moeda de 5 cntimos para
     *    outro saco com uma moeda de 1 cntimo, transfere-se a moeda de 
     *    5 cntimos num sentido e a moeda de 1 cntimo de troco em sentido 
     *    inverso.
     */
    public boolean transferir(int montante, SacoMoedas destino)
    {
       if (montante == 0)
            return true;
           
       if (getMontante() < montante)
            return false;
      
      // Cria saco auxiliar "destinoAux", com contedo igual ao saco de destino 
      // para procurar solues. Cada vez que encontra uma soluo melhor (com
      // menos moedas) copia a soluo para "destino".
      SacoMoedas destinoAux = new SacoMoedas();
      SacoMoedas destinoBest = new SacoMoedas();
      transferir(montante, destinoAux, moedas.length - 1, destinoBest);
      if (destinoBest.getMontante() == montante)
      {
            this.subtrair(destinoBest);
            destino.somar(destinoBest);
            return true;
      }
            
      System.out.println("falhou: origem=" + this + " destino=" + destino);
       // Se chegou aqui, no  possivel transferir apenas num sentido.
       // Assim, tenta outra estratgia com transferncias nos dois sentidos.

       // - memoriza montante inicial no saco de destino
       int montanteDestino = destino.getMontante();

       // - transfere todas as moedas do saco de destino para o saco de origem 
       //  (tem de suceder sempre)
       destino.transferir(montanteDestino, this);
 
       System.out.println("despois de juntar: origem=" + this + " destino=" + destino);
  
       // - tenta agora transferir para o saco de destino o montante necessrio
       if (transferir(montanteDestino + montante, destino))
          return true;
          
       // - se falhou, tem de repor montante anterior (tem de suceder sempre)
       transferir(montanteDestino, destino);
       return false;
    } 
    
    /**
     * Mtodo auxiliar usado pelo mtodo anterior. Tem um argumento adicional
     * que indica o ndice mximo das moedas que pode transferir.
     */
    public void transferir(int montante, SacoMoedas destino, int indiceMax, SacoMoedas destinoBest)
    {
       if (montante == 0)
       {
         // se  a primeira soluo encontrada ou tem menos moedas, guarda
         if (destino.getMontante() > destinoBest.getMontante() // 1 soluo
            || destino.getNumMoedas() < destinoBest.getNumMoedas())
                destino.copiarPara(destinoBest);
         return;
       }
                            
        for (int i = indiceMax; i >= 0; i--)
        {
           if (quant[i] > 0 && moedas[i] <= montante 
               && destino.aceitaMoeda(moedas[i]))
           {
               // transfere uma moeda
               quant[i]--;
               destino.quant[i]++;
               montante -= moedas[i];
    
               // transfere o valor restante recursivamente
               // (no vale a pena usar moedas mais altas que esta ...)
               transferir(montante,destino, i, destinoBest);

               // repe situao anterior para procurar outras solues
               quant[i]++;
               destino.quant[i]--;
               montante += moedas[i];
           }
        }
        
    } 

    /**
     * Mtodo auxiliar que retorna o nmero de moedas existentes no saco.
     */
    private int getNumMoedas()
    {
        int sum = 0;
        for (int i = moedas.length-1; i >= 0; i--)
            sum += quant[i];
        return sum;
    }
    
    /**
     * Mtodo auxliar que copia o contedo deste saco para outro saco.
     */
    private void copiarPara(SacoMoedas destino)
    {
        for (int i = moedas.length-1; i >= 0; i--)
            destino.quant[i] = quant[i];
    }
    
    /**
     * Mtodo auxliar que soma a este saco as quantidades existentes num saco2,
     * sem alterar o saco2.
     */
    private void somar(SacoMoedas saco2)
    {
        for (int i = moedas.length-1; i >= 0; i--)
            quant[i] += saco2.quant[i];
    }
    /**
     * Mtodo auxliar que subtrai a este saco as quantidades existentes num saco2,
     * sem alterar o saco2.
     */
    private void subtrair(SacoMoedas saco2)
    {
        for (int i = moedas.length-1; i >= 0; i--)
            quant[i] -= saco2.quant[i];
    }
   
    /**
     * Mtodo "main" para testar esta classe.
     */
    public static void main(String args[])
    {
        // Cria um saco1 com quatro moedas de 20 cntimos, uma moeda de 50
        // cntimos e duas moedas de 1 euro
        SacoMoedas saco1 = new SacoMoedas();
        saco1.setQuant(20, 4);
        saco1.setQuant(50, 1);
        saco1.setQuant(100, 2);
        
        // Imprime o contedo e montante total do saco1 no ecr
        System.out.println("Saco1: "+saco1+" total: "+saco1.getMontante());
        
        // Cria um saco2 vazio
        SacoMoedas saco2 = new SacoMoedas();
        
        // Tenta transferir 260 cntimos do saco1 para o saco2
        saco1.transferir(260,saco2);

       // Volta a imprimir o contedo e montante total do saco1 no ecr
        System.out.println("Saco1: "+saco1+" total: "+saco1.getMontante());
 
        // Imprime o contedo e montante total do saco2 no ecr
        System.out.println("Saco2: "+saco2+" total: "+saco2.getMontante());
    }
    
    /**
     * Mtodo de teste para testar o primeiro melhoramento do mtodo 
     * "transferir" (minimizar nmero de moedas transferidas).
     */
    public static void Teste2()
    {
        // Cria um saco1 com as moedas {2, 2, 2, 2, 2, 20, 20, 20, 50} 
        SacoMoedas saco1 = new SacoMoedas();
        saco1.setQuant(2, 5);
        saco1.setQuant(20, 3);
        saco1.setQuant(50, 1);

        // Cria um saco2 vazio
        SacoMoedas saco2 = new SacoMoedas();
        
        // Tenta transferir 60 cntimos do saco1 para o saco2
        saco1.transferir(60, saco2);
        
        // Imprime o contedo e montante total dos dois sacos
        System.out.println("Saco1: "+saco1+" total: "+saco1.getMontante());
        System.out.println("Saco2: "+saco2+" total: "+saco2.getMontante());
    }
    
    /**
     * Mtodo de teste para testar o segundo melhoramento do mtodo 
     * "transferir" (com transferencias nos dois sentidos).
     */
    public static void Teste3()
    {
        // Cria um saco1 com uma moeda de 5 cntimos
        SacoMoedas saco1 = new SacoMoedas();
        saco1.setQuant(5, 1);
         
        // Cria um saco2 com uma moeda de 1 cntimo
        SacoMoedas saco2 = new SacoMoedas();
        saco2.setQuant(1, 1);
        
        // Tenta transferir 4 cntimos do saco1 para o saco2
        saco1.transferir(4, saco2);
        
        // Imprime o contedo e montante total dos dois sacos
        System.out.println("Saco1: "+saco1+" total: "+saco1.getMontante());
        System.out.println("Saco2: "+saco2+" total: "+saco2.getMontante());
    }


}
